const std = @import("std");

pub const PackError = error{InvalidPack};

pub const Entry = struct {
    size: u32,
    offset: u32,
    value: u32,
    oname: []const u8,
    name: []const u8,
};

allocator: std.mem.Allocator,
handle: std.fs.File,
description: []const u8,
entries: []Entry,

pub fn open(allocator: std.mem.Allocator, name: []const u8) !@This() {
    const f = try std.fs.cwd().openFile(name, .{});
    const reader = f.reader();

    const magic = try reader.readInt(u32, .big);
    if (magic != 0x5041434b) return PackError.InvalidPack;

    const size = try reader.readInt(u32, .little);
    const desc = try reader.readBoundedBytes(32);
    const a: []const u8 = try allocator.dupe(u8, &desc.buffer);
    const d = std.mem.trimRight(u8, a, "\x00");
    const entries = try allocator.alloc(Entry, size);

    var i: u32 = 0;
    while (i < size) : (i += 1) {
        const fsize = try reader.readInt(u32, .little);
        const foffset = try reader.readInt(u32, .little);
        const fvalue = try reader.readInt(u32, .little);
        const foname = try reader.readUntilDelimiterAlloc(allocator, 0, 64);
        const fname = try allocator.dupe(u8, foname);
        std.mem.replaceScalar(u8, fname, std.fs.path.sep_windows, std.fs.path.sep);
        entries[i] = .{
            .size = fsize,
            .offset = foffset,
            .value = fvalue,
            .oname = foname,
            .name = fname,
        };
    }

    return .{
        .allocator = allocator,
        .handle = f,
        .description = d,
        .entries = entries,
    };
}

pub fn close(self: @This()) void {
    for (self.entries) |entry| {
        self.allocator.free(entry.name);
    }
    self.allocator.free(self.entries);
    self.handle.close();
}

pub fn copy(self: @This(), entry: Entry, out: std.fs.File, offset: u64) !u64 {
    return self.handle.copyRangeAll(entry.offset, out, offset, entry.size);
}
